import numpy as np
from yupi.affine_estimator import affine_matrix

def add_dynamic_reference(trajectory, reference, start_in_origin=True):
    """ 
    This functions fuse the information of a trajectory with an 
    external reference of the motion of the System of Reference
    (SoR).

    It allows to remap the information gathered in local SoRs
    to a more general SoR.
    """
    def affine2camera(theta, tx, ty):
        x_cl, y_cl, theta_cl = np.zeros((3, theta.size + 1))
        theta_cl[1:] = np.cumsum(theta)

        for i in range(theta.size):
            A = affine_matrix(theta_cl[i + 1], x_cl[i], y_cl[i], R_inv=True)
            x_cl[i + 1], y_cl[i + 1] = A @ [-tx[i], -ty[i], 1]

        x_cl, y_cl, theta_cl = x_cl[1:], y_cl[1:], theta_cl[1:]
        return x_cl, y_cl, theta_cl


    def camera2obj(x_ac, y_ac, x_cl, y_cl, theta_cl):
        x_al, y_al = np.empty((2, x_ac.size))

        for i in range(x_ac.size):
            A = affine_matrix(theta_cl[i], x_cl[i], y_cl[i], R_inv=True)
            x_al[i], y_al[i] = A @ [x_ac[i], y_ac[i], 1]

        return x_al, y_al


    def affine2obj(theta, tx, ty, x_ac, y_ac):
        x_cl, y_cl, theta_cl = affine2camera(theta, tx, ty)
        x_al, y_al = camera2obj(x_ac, y_ac, x_cl, y_cl, theta_cl)
        return x_al, y_al

    theta, tx, ty = reference

    x_al, y_al = affine2obj(theta, tx, ty, trajectory.x_arr, trajectory.y_arr)
    
    if start_in_origin:
        x_al = x_al - x_al[0]
        y_al = y_al - y_al[0]

    trajectory.x_arr = x_al
    trajectory.y_arr = y_al

    return trajectory


import numpy as np



class RandomWalk:
    """
    Multidimensional Random Walk class.
    """
    
    def __init__(self, n:int, dim:int=1, N:int=1, dt:float=1, 
            actions:np.ndarray=None, actions_prob:np.ndarray=None, 
            jump_len:np.ndarray=None):

        # simulation parameters
        self.n = n      # total number of steps
        self.dim = dim  # random walk dimension
        self.N = N      # number of walkers
        self.dt = dt    # time step
        
        # model parameters
        self.actions = actions            # available RW movements
        self.actions_prob = actions_prob  # probability for every action
        self.jump_len = jump_len          # statistic for jumps length

        # dynamic variables
        self.t = np.arange(n) * dt      # time array
        self.r = np.zeros((n, dim, N))  # position array

        # only right/left jumps, uniform probabilities for it
        # and equal length for all jumps are set as default
        if actions is None:
            self.actions = np.array([1, -1])
        if actions_prob is None:
            self.actions_prob = np.tile([.5, .5], (dim, 1))
        if jump_len is None:
            self.jump_len = np.ones((dim, N))


    # compute vector position as a function of time for
    # all the walkers of the ensemble
    def get_r(self):
        # get movements for every space coordinates according 
        # to the sample space of probabilities in self.actions_prob
        dr = [np.random.choice(actions, p=p, size=(self.n - 1, self.N)) for p in self.actions_prob]
        
        # set time/coordinates as the first/second axis
        dr = np.swapaxes(dr, 0, 1)
        
        # scale displacements according to the jump length statistics
        dr = dr * self.jump_len

        # integrate displacements to get position vectors
        self.r[1:] = np.cumsum(dr, axis=0)
        return self.r



# testing
if __name__ == '__main__':

    import matplotlib.pyplot as plt

    np.random.seed(0)

    # set parameter values
    n = 500
    dim = 2
    N = 5
    dt = 1
    actions = [1, 0, -1]
    prob = [[.5, .1, .4],
            [.5, 0, .5]]

    # get RandomWalk object and get position vectors
    rw = RandomWalk(n, dim, N, dt, actions, prob)
    r = rw.get_r()
    x, y = r[:,0,:], r[:,1,:]

    # plotting
    plt.plot(x, y)
    plt.axis('equal')
    plt.xlabel('x')
    plt.ylabel('y')
    plt.show()


import numpy as np
from yupi import Trajectory


class LE:
	"""
	Random Walk class from a multidimensional Langevin Equation.
	"""

	def __init__(self, T:float, tau:float, noise_params:dict, 
				dim:int=1, N:int=1, dt:int=1):

		# siulation parameters
		self.T = T                    # total time
		self.dt = dt                  # time step of the simulation
		self.N = N                    # number of trials
		self.n = int(T / dt)          # number of time steps
		self.size = (self.n, dim, N)  # shape of the dynamic variables

		# model parameters
		self.tau = tau                    # relaxation characteristic time
		self.noise_params = noise_params  # noise properties (keys: 'pdf_name', 'scale')
		self.noise = np.ndarray           # noise array that will be fill in get_noise method

		# dynamic variables
		self.t = np.linspace(0, T, num=self.n)  # time array
		self.r = np.empty(self.size)            # position array
		self.v = np.empty(self.size)            # velocity array

		# initial conditions
		self.r0 = np.zeros((dim, self.N))
		self.v0 = np.zeros((dim, self.N))
		self._r0_ = False                  # True if initial positions are set by the user
		self._v0_ = False                  # True if initial velocities are set by the user


	# set initial condition for position vectors
	def set_r_init_cond(self, r0=None):
		if r0 is None:
			self.r[0] = self.r0  # set as default
		else:
			self.r[0] = r0       # set by the user

		self._r0_ = True


	# set initial condition for velocity vectors
	def set_v_init_cond(self, v0=None):
		self.v[0] = self.v0 if v0 is None else v0 # set as default
		self._v0_ = True


	# fill noise array with custom noise properties
	def get_noise(self):
		pdf_name = self.noise_params['pdf_name']  # noise PDF
		scale = self.noise_params['scale']        # scale parameter (not stan. dev.)

		dist = getattr(np.random, pdf_name)
		noise = dist(scale=scale, size=self.size)

		self.noise = noise


	# solve Langevin Equation using the numerical method of Euler-Maruyama
	def solve_rv(self):
		for i in range(self.n - 1):
			# solving for position
			self.r[i + 1] = self.r[i] + \
							self.v[i] * self.dt

			# solving for velocity
			self.v[i + 1] = self.v[i] + \
							-np.dot(1 / self.tau, self.v[i]) * self.dt + \
							self.noise[i] * np.sqrt(self.dt)


	# simulate the process
	def simulate(self):
		# set initial conditions
		if not self._r0_:
			self.set_r_init_cond()
		if not self._v0_:
			self.set_v_init_cond()

		self.get_noise()  # create the attribute self.noise
		self.solve_rv()   # solve the Langevin equation

		x, y = self.r[:,0,:], self.r[:,1,:]
		# return self.r, self.v
		return Trajectory(x_arr=x, y_arr=y, dt=self.dt,
                                  id="LangevinSolution")



# testing
if __name__ == '__main__':
	from yupi.analyzing.visualization import plot_trajectories
	
	np.random.seed(0)

	# set parameter values
	dim = 2
	T = 5 * 60
	dt = .05
	N = 5
	tau = 1.	

	kTm = 10
	scale = np.sqrt(2 * kTm / tau)
	noise_params = {'pdf_name': 'normal', 'scale': scale}

	# get LE object and get position vectors
	le = LE(T, tau, noise_params, dim, N, dt)
	le.set_v_init_cond(scale * np.random.randn(le.N))

	tr = le.simulate()
	plot_trajectories([tr])



import abc
import cv2

def resize_frame(frame, scale=1):
    h, w = frame.shape[:2]
    w_, h_ = int(scale * w), int(scale * h)
    short_frame = cv2.resize(frame, (w_, h_), interpolation=cv2.INTER_AREA)
    return short_frame


def change_colorspace(image, color_space):
    if color_space == 'BGR':
        return image.copy()
    if color_space == 'GRAY':
        return cv2.cvtColor(image.copy(), cv2.COLOR_BGR2GRAY)
    if color_space == 'HSV':
        return cv2.cvtColor(image.copy(), cv2.COLOR_BGR2HSV)

        
class Algorithm(metaclass=abc.ABCMeta):
    """docstring for Algorithm"""
    def __init__(self):
        pass
        
    def get_centroid(self, bin_img):
        # Calculate moments
        M = cv2.moments(bin_img)  

        # Check if something was over the threshold
        if M['m00'] != 0:
            # calculate x,y coordinate of center
            cX = int(M['m10'] / M['m00'])
            cY = int(M['m01'] / M['m00'])
            return cX, cY
        else:
            print('[ERROR] Nothing was over threshold')
            return 0, 0

    @abc.abstractmethod
    def detect(self, current_chunk, previous_chunk):
        pass

# TODO: Fix this to emulate the previous algorithm including this:
# ant_ratio = ant_pixels / (roi_width * roi_heigh) # approximate ratio of the ant compare to the roi
class IntensityMatching(Algorithm):
    """docstring for IntensityMatching"""
    def __init__(self, min_val=0, max_val=255, max_pixels=None):
        super(IntensityMatching, self).__init__()
        self.min_val = min_val
        self.max_val = max_val
        self.max_pixels = max_pixels

    def detect(self, current_chunk, previous_chunk=None):
        # convert image to grayscale image
        gray_image = cv2.cvtColor(current_chunk.copy(), cv2.COLOR_BGR2GRAY)

        if self.max_pixels:
            # obtain image histogram
            ys = cv2.calcHist([gray_image], [0], None, [256], [0,256], accumulate=True)

            # compute image total pixel count
            x, y = gray_image.shape
            total = x * y

            # compute an adaptative threshold according the ratio of darkest pixels
            max_threshold = self.max_val
            for i in range(self.min_val, self.max_val):
                if ys[i] > self.max_pixels:
                    max_threshold = i
                    break
            self.max_val = max_threshold

        mask = cv2.inRange(gray_image, self.min_val, self.max_val) 
        centroid = self.get_centroid(mask)
        # convert the grayscale image to binary image
        return mask, centroid


class ColorMatching(Algorithm):
    """docstring for ColorMatching"""
    def __init__(self, lower_bound=(0,0,0), upper_bound=(255,255,255), 
        color_space='BGR', max_pixels=None):
        super(ColorMatching, self).__init__()
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.color_space = color_space
        self.max_pixels = max_pixels

    def detect(self, current_chunk, previous_chunk=None):

        # convert image to desired colorspace
        copied_image = change_colorspace(current_chunk, self.color_space)

        mask = cv2.inRange(copied_image, self.lower_bound, self.upper_bound) 
        centroid = self.get_centroid(mask)
        # convert the grayscale image to binary image
        return mask, centroid
# TODO: Convert this to Algorithm Class

# def frame_diff_detector(frame1, frame2):
#     # obtain the frame difference
#     diff = cv2.absdiff(frame1, frame2)

#     # convert image to grayscale image
#     gray_image = cv2.cvtColor(diff, cv2.COLOR_BGR2GRAY)

#     # convert the grayscale image to binary image
#     ret, thresh = cv2.threshold(gray_image, sett.frame_diff_threshold, 255, cv2.THRESH_BINARY)

#     # give a hint message about what may go wrong
#     hint = "Try decreasing the value of settings.frame_diff_threshold"
    
#     # return the centroid of the pixels over threshold
#     return get_centroid(thresh, hint)





"""
Tracking package docstring
"""

from yupi.tracking.trackers import (
    ROI,
    ObjectTracker,
    CameraTracker,
    TrackingScenario
)
from yupi.tracking.algorithms import (
    Algorithm,
    IntensityMatching,
    ColorMatching
)
from yupi.tracking.undistorters import (
    Undistorter,
    ClassicUndistorter,
    RemapUndistorter,
    NoUndistorter
)

__all__ = [
    'ROI',
    'ObjectTracker',
    'CameraTracker',
    'TrackingScenario',
    'Algorithm',
    'IntensityMatching',
    'ColorMatching',
    'Undistorter',
    'ClassicUndistorter',
    'RemapUndistorter',
    'NoUndistorter'
]


"""
All package docstring
"""

from yupi.trajectory import Trajectory
import logging

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s] %(message)s',
    datefmt='%H:%M:%S'
)

__all__ = [
    'Trajectory'
]

__version__ = '0.2.2'